import 'dart:async';

import 'package:talker/talker.dart';

class Talker implements TalkerInterface {
  Talker._() {
    _settings = kDefaultTalkerSettings;
    _logger = TalkerLogger();
    _errorHandler = ErrorHandler()
      ..stream.listen((details) {
        _handleErrorStream(details);
      });
  }

  static final _talker = Talker._();
  static TalkerInterface get instance => _talker;

  /// Fields can be setup in [configure()] method
  late TalkerSettings _settings;
  late TalkerLogger _logger;
  late ErrorHandler _errorHandler;

  // final _fileManager = FileManager();
  final _history = <TalkerDataInterface>[];
  TalkerObserversManager? _observersManager;

  @override
  Future<void> configure({
    TalkerLogger? logger,
    ErrorHandler? errorHandler,
    TalkerSettings? settings,
    List<TalkerObserver>? observers,
  }) async {
    if (settings != null) {
      _settings = settings;
    }

    if (observers != null && observers.isNotEmpty) {
      _observersManager = TalkerObserversManager(observers);
    }

    if (logger != null) {
      _logger = logger;
    }

    if (errorHandler != null) {
      _errorHandler = errorHandler
        ..stream.listen((details) {
          _handleErrorStream(details);
        });
    }
  }

  final _talkerStreamController =
      StreamController<TalkerDataInterface>.broadcast();

  @override
  Stream<TalkerDataInterface> get stream =>
      _talkerStreamController.stream.asBroadcastStream();

  @override
  List<TalkerDataInterface> get history => _history;

  @override
  void handle(
    Object exception, [
    String? msg,
    StackTrace? stackTrace,
    ErrorLevel? errorLevel,
  ]) {
    final details = _errorHandler.handle(
      exception,
      msg,
      stackTrace,
      errorLevel,
    );
    if (details != null) {
      _observersManager?.onError(details);
    }
  }

  @override
  void handleError(
    Error error, [
    String? msg,
    StackTrace? stackTrace,
    ErrorLevel? errorLevel,
  ]) {
    final errContainer =
        _errorHandler.handleError(error, msg, stackTrace, errorLevel);
    _observersManager?.onError(errContainer);
  }

  @override
  void handleException(
    Exception exception, [
    String? msg,
    StackTrace? stackTrace,
    ErrorLevel? errorLevel,
  ]) {
    final errContainer =
        _errorHandler.handleException(exception, msg, stackTrace, errorLevel);
    _observersManager?.onError(errContainer);
  }

  @override
  void log(
    String message, {
    LogLevel logLevel = LogLevel.debug,
    Map<String, dynamic>? additional,
    Object? exception,
    StackTrace? stackTrace,
    AnsiPen? pen,
  }) {
    _handleLog(
      exception,
      message,
      stackTrace,
      logLevel,
      additional: additional,
      pen: pen,
    );
  }

  @override
  void critical(
    String msg, [
    Object? exception,
    StackTrace? stackTrace,
  ]) {
    _handleLog(exception, msg, stackTrace, LogLevel.critical);
  }

  @override
  void debug(
    String msg, [
    Object? exception,
    StackTrace? stackTrace,
  ]) {
    _handleLog(exception, msg, stackTrace, LogLevel.debug);
  }

  @override
  void error(
    String msg, [
    Object? exception,
    StackTrace? stackTrace,
  ]) {
    _handleLog(exception, msg, stackTrace, LogLevel.error);
  }

  @override
  void fine(
    String msg, [
    Object? exception,
    StackTrace? stackTrace,
  ]) {
    _handleLog(exception, msg, stackTrace, LogLevel.fine);
  }

  @override
  void good(
    String msg, [
    Object? exception,
    StackTrace? stackTrace,
  ]) {
    _handleLog(exception, msg, stackTrace, LogLevel.good);
  }

  @override
  void info(
    String msg, [
    Object? exception,
    StackTrace? stackTrace,
  ]) {
    _handleLog(exception, msg, stackTrace, LogLevel.info);
  }

  @override
  void verbose(
    String msg, [
    Object? exception,
    StackTrace? stackTrace,
  ]) {
    _handleLog(exception, msg, stackTrace, LogLevel.verbose);
  }

  @override
  void warning(
    String msg, [
    Object? exception,
    StackTrace? stackTrace,
  ]) {
    _handleLog(exception, msg, stackTrace, LogLevel.warning);
  }

  void _handleLog(
    Object? exception,
    String message,
    StackTrace? stackTrace,
    LogLevel logLevel, {
    Map<String, dynamic>? additional,
    AnsiPen? pen,
  }) {
    TalkerDataInterface? data;

    if (exception != null) {
      handle(exception, message, stackTrace);
      return;
    }

    data = TalkerLog(
      message,
      logLevel: logLevel,
      additional: additional,
    );
    _talkerStreamController.add(data);
    _observersManager?.onLog(data);
    _handleForOutputs(data);
    if (_settings.useConsoleLogs) {
      _logger.log(
        data.generateTextMessage(),
        level: data.logLevel,
        pen: pen,
      );
    }
  }

  @override
  void cleanHistory() {
    if (_settings.useHistory) {
      _history.clear();
    }
  }

  void _handleForOutputs(TalkerDataInterface data) {
    if (_settings.useHistory) {
      _writeToHistory(data);
    }
    // _writeToFile(data);
  }

  //TODO: recreate file manager logic
  // void _writeToFile(TalkerDataInterface data) {
  //   if (_settings.writeToFile) {
  //     _fileManager.writeToLogFile(data.generateTextMessage());
  //   }
  // }

  void _writeToHistory(TalkerDataInterface data) {
    if (_settings.useHistory) {
      if (_settings.maxHistoryItems <= _history.length) {
        _history.removeAt(0);
      }
      _history.add(data);
    }
  }

  void _handleErrorStream(ErrorDetails details) {
    TalkerDataInterface? data;
    final err = details.error;
    final exception = details.exception;
    if (err != null) {
      data = TalkerError(
        err,
        message: details.message,
        stackTrace: details.stackTrace,
        logLevel: details.errorLevel?.loglevel ?? LogLevel.error,
      );
    } else if (exception != null) {
      data = TalkerException(
        exception,
        message: details.message,
        stackTrace: details.stackTrace,
        logLevel: details.errorLevel?.loglevel ?? LogLevel.error,
      );
    }

    if (data != null) {
      _talkerStreamController.add(data);
      _handleForOutputs(data);
      if (_settings.useConsoleLogs) {
        _logger.log(
          data.generateTextMessage(),
          level: data.logLevel ?? LogLevel.error,
        );
      }
    }
  }
}
